import logging

from django.conf import settings

app_logger = logging.getLogger("app")


class ExcludeCommonURLsFilter(logging.Filter):
    def __init__(self):
        self.exclude_urls = [
            "/admin/jsi18n/",  # could be customized...
            "/favicon.ico",
        ]

        if settings.STATIC_URL:
            self.exclude_urls.append(settings.STATIC_URL)

        super().__init__()

    def filter(self, record):
        message = record.getMessage()

        if message.startswith('"GET /'):  # Only check them all if it's a GET
            for url in self.exclude_urls:
                if message.startswith(f'"GET {url}'):
                    return False

        return True


class KVLogger:
    def __init__(self, logger):
        self.logger = logger

    def log(self, level, message, **kwargs):
        self.logger.log(level, f"{message} {self._format_kwargs(kwargs)}")

    def _format_kwargs(self, kwargs):
        outputs = []

        for k, v in kwargs.items():
            self._validate_key(k)
            formatted_value = self._format_value(v)
            outputs.append(f"{k}={formatted_value}")

        return " ".join(outputs)

    def _validate_key(self, key):
        if " " in key:
            raise ValueError("Keys cannot have spaces")

        if "=" in key:
            raise ValueError("Keys cannot have equals signs")

        if '"' in key or "'" in key:
            raise ValueError("Keys cannot have quotes")

    def _format_value(self, value):
        if isinstance(value, str):
            s = value
        else:
            s = str(value)

        if '"' in s:
            # Escape quotes and surround it
            s = s.replace('"', '\\"')
            s = f'"{s}"'
        elif s == "":
            # Quote empty strings instead of printing nothing
            s = '""'
        elif any(char in s for char in [" ", "/", "'", ":", "=", "."]):
            # Surround these with quotes for parsers
            s = f'"{s}"'

        return s

    def info(self, message, **kwargs):
        self.log(logging.INFO, message, **kwargs)

    def debug(self, message, **kwargs):
        self.log(logging.DEBUG, message, **kwargs)

    def warning(self, message, **kwargs):
        self.log(logging.WARNING, message, **kwargs)

    def error(self, message, **kwargs):
        self.log(logging.ERROR, message, **kwargs)

    def critical(self, message, **kwargs):
        self.log(logging.CRITICAL, message, **kwargs)


# Make this accessible from the app_logger
app_logger.kv = KVLogger(app_logger)
